import builtins
import enum
import functools
import operator
import tomllib
import types
from importlib import resources
from typing import *

import setdoc

__all__ = ["Scaevola", "auto", "getfuncnames", "makefunc"]


class Util(enum.Enum):
    "This enum provides a singleton."

    util = None

    @functools.cached_property
    def data(self: Self) -> dict:
        "This cached property holds the cfg data."
        text: str
        text = resources.read_text("scaevola.core", "cfg.toml")
        return tomllib.loads(text)


def auto(cls: type) -> type:
    "This decorator implements all the righthand functions."
    name: str
    for name in getfuncnames():
        if name not in cls.__dict__.keys():
            makefunc(cls, name)
    return cls


def getfuncnames() -> list[str]:
    "This function returns the names of all righthand functions."
    return list(Util.util.data.keys())


def makefunc(cls: type, name: str) -> types.FunctionType:
    "This function implements a certain righthand function."
    funcname: str
    inner: Callable
    module: Any
    funcname = Util.util.data[name]["func"]
    if Util.util.data[name].get("isbuiltin", False):
        module = builtins
    else:
        module = operator
    inner = getattr(module, funcname)

    def outer(self: Self, other: Any) -> Any:
        "This docstring will be overwritten."
        return inner(type(self)(other), self)

    outer.__module__ = cls.__module__
    outer.__name__ = name
    outer.__qualname__ = cls.__qualname__ + "." + name
    setdoc.basic(outer)
    setattr(cls, name, outer)
    return outer


@auto
class Scaevola:
    __slots__ = ()



import unittest
from typing import *

from scaevola.core import Scaevola

__all__ = ["TestScaevola"]


# Subclass of Scaevola to implement basic arithmetic and comparison operations
class ScaevolaSubclass(Scaevola):
    def __init__(self: Self, value: Any) -> Any:
        self.value = value

    def __eq__(self: Self, other: Any) -> Any:
        if isinstance(other, ScaevolaSubclass):
            return self.value == other.value
        raise NotImplementedError

    def __lt__(self: Self, other: Any) -> Any:
        if isinstance(other, ScaevolaSubclass):
            return self.value < other.value
        raise NotImplementedError

    def __le__(self: Self, other: Any) -> Any:
        if isinstance(other, ScaevolaSubclass):
            return self.value <= other.value
        raise NotImplementedError

    def __add__(self: Self, other: Any) -> Any:
        if isinstance(other, ScaevolaSubclass):
            return ScaevolaSubclass(self.value + other.value)
        raise NotImplementedError

    def __and__(self: Self, other: Any) -> Any:
        if isinstance(other, ScaevolaSubclass):
            return ScaevolaSubclass(self.value & other.value)
        raise NotImplementedError

    def __divmod__(self: Self, other: Any) -> Any:
        if isinstance(other, ScaevolaSubclass):
            return divmod(self.value, other.value)
        raise NotImplementedError

    def __floordiv__(self: Self, other: Any) -> Any:
        if isinstance(other, ScaevolaSubclass):
            return ScaevolaSubclass(self.value // other.value)
        raise NotImplementedError

    def __lshift__(self: Self, other: Any) -> Any:
        if isinstance(other, ScaevolaSubclass):
            return ScaevolaSubclass(self.value << other.value)
        raise NotImplementedError

    def __matmul__(self: Self, other: Any) -> Any:
        if isinstance(other, ScaevolaSubclass):
            return ScaevolaSubclass(
                self.value * other.value
            )  # Simulating matrix multiplication
        raise NotImplementedError

    def __mod__(self: Self, other: Any) -> Any:
        if isinstance(other, ScaevolaSubclass):
            return ScaevolaSubclass(self.value % other.value)
        raise NotImplementedError

    def __mul__(self: Self, other: Any) -> Any:
        if isinstance(other, ScaevolaSubclass):
            return ScaevolaSubclass(self.value * other.value)
        raise NotImplementedError

    def __or__(self: Self, other: Any) -> Any:
        if isinstance(other, ScaevolaSubclass):
            return ScaevolaSubclass(self.value | other.value)
        raise NotImplementedError

    def __pow__(self: Self, other: Any) -> Any:
        if isinstance(other, ScaevolaSubclass):
            return ScaevolaSubclass(self.value**other.value)
        raise NotImplementedError

    def __rshift__(self: Self, other: Any) -> Any:
        if isinstance(other, ScaevolaSubclass):
            return ScaevolaSubclass(self.value >> other.value)
        raise NotImplementedError

    def __sub__(self: Self, other: Any) -> Any:
        if isinstance(other, ScaevolaSubclass):
            return ScaevolaSubclass(self.value - other.value)
        raise NotImplementedError

    def __truediv__(self: Self, other: Any) -> Any:
        if isinstance(other, ScaevolaSubclass):
            return ScaevolaSubclass(self.value / other.value)
        raise NotImplementedError

    def __xor__(self: Self, other: Any) -> Any:
        if isinstance(other, ScaevolaSubclass):
            return ScaevolaSubclass(self.value ^ other.value)
        raise NotImplementedError


class TestScaevola(unittest.TestCase):
    def setUp(self: Self) -> None:
        self.obj1 = ScaevolaSubclass(10)
        self.obj2 = ScaevolaSubclass(20)

    def test_ge(self: Self) -> None:
        self.assertTrue(self.obj2 >= 10)
        self.assertFalse(self.obj1 >= 20)

    def test_gt(self: Self) -> None:
        self.assertTrue(self.obj2 > 10)
        self.assertFalse(self.obj1 > 20)

    def test_radd(self: Self) -> None:
        result = 15 + self.obj1
        self.assertEqual(result.value, 25)

    def test_rand(self: Self) -> None:
        result = 15 & self.obj1
        self.assertEqual(result.value, 10 & 15)

    def test_rdivmod(self: Self) -> None:
        result = divmod(25, self.obj1)
        self.assertEqual(result, (2, 5))

    def test_rfloordiv(self: Self) -> None:
        result = 25 // self.obj1
        self.assertEqual(result.value, 2)

    def test_rlshift(self: Self) -> None:
        result = 2 << self.obj1
        self.assertEqual(result.value, 2 << 10)

    def test_rmatmul(self: Self) -> None:
        result = 2 @ self.obj1
        self.assertEqual(result.value, 2 * 10)

    def test_rmod(self: Self) -> None:
        result = 23 % self.obj1
        self.assertEqual(result.value, 23 % 10)

    def test_rmul(self: Self) -> None:
        result = 2 * self.obj1
        self.assertEqual(result.value, 20)

    def test_ror(self: Self) -> None:
        result = 2 | self.obj1
        self.assertEqual(result.value, 2 | 10)

    def test_rpow(self: Self) -> None:
        result = 2**self.obj1
        self.assertEqual(result.value, 2**10)

    def test_rrshift(self: Self) -> None:
        result = 1024 >> self.obj1
        self.assertEqual(result.value, 1024 >> 10)

    def test_rsub(self: Self) -> None:
        result = 25 - self.obj1
        self.assertEqual(result.value, 25 - 10)

    def test_rtruediv(self: Self) -> None:
        result = 100 / self.obj1
        self.assertEqual(result.value, 100 / 10)

    def test_rxor(self: Self) -> None:
        result = 15 ^ self.obj1
        self.assertEqual(result.value, 15 ^ 10)


if __name__ == "__main__":
    unittest.main()



import unittest
from typing import *

from scaevola.core import auto

__all__ = ["TestSlots"]


@auto
class Foo:
    __slots__ = ("x", "y")


@auto
class Bar:
    pass


class TestSlots(unittest.TestCase):
    def test_foo(self: Self) -> None:
        foo: Foo
        foo = Foo()
        foo.x = 4
        foo.y = 2
        self.assertEqual(foo.x, 4)
        self.assertEqual(foo.y, 2)
        with self.assertRaises(AttributeError):
            foo.z = 0

    def test_bar(self: Self) -> None:
        bar: Bar
        bar = Bar()
        bar.x = 4
        bar.y = 2
        self.assertEqual(bar.x, 4)
        self.assertEqual(bar.y, 2)


if __name__ == "__main__":
    unittest.main()



import unittest

__all__ = ["test"]


def test() -> unittest.TextTestResult:
    loader: unittest.TestLoader
    suite: unittest.TestSuite
    loader = unittest.TestLoader()
    suite = loader.discover(start_dir="scaevola.tests")
    return unittest.TextTestRunner().run(suite)



import unittest
from typing import *

from scaevola.core import Scaevola

__all__ = ["TestScaevolaDocstrings"]


class TestScaevolaDocstrings(unittest.TestCase):
    def test_methods_have_docstrings(self: Self) -> None:
        magic: Any
        methods_to_check: list[str]
        name: str
        methods_to_check = [
            "__ge__",
            "__gt__",
            "__radd__",
            "__rand__",
            "__rdivmod__",
            "__rfloordiv__",
            "__rlshift__",
            "__rmatmul__",
            "__rmod__",
            "__rmul__",
            "__ror__",
            "__rpow__",
            "__rrshift__",
            "__rsub__",
            "__rtruediv__",
            "__rxor__",
        ]
        for name in methods_to_check:
            with self.subTest(method=name):
                magic = getattr(Scaevola, name, None)
                self.assertTrue(
                    magic.__doc__.startswith("This magic method "),
                    "doc=%r" % magic.__doc__,
                )


if __name__ == "__main__":
    unittest.main()
